"""Helper functions for writing test suites."""

from typing import Any, Callable, List, Protocol, Tuple, TypeVar, Union, cast

import pytest
from outcome.utils import env


class Mark(Protocol):
    name: str
    args: Tuple[Any, ...]


class Marked(Protocol):
    pytestmark: List[Mark]


# Taken from pytest
_Markable = TypeVar('_Markable', bound=Union[Callable[..., object], type])

# We cast to _Markable as pytest uses type overrides which obscure the real time


def skip_for_e2e(fn: _Markable) -> _Markable:
    decorated = pytest.mark.skipif(env.is_e2e(), reason='Skipped in e2e tests')(fn)
    return cast(_Markable, decorated)


def only_for_e2e(fn: _Markable) -> _Markable:
    decorated = pytest.mark.skipif(not env.is_e2e(), reason='Only for e2e tests')(fn)
    return cast(_Markable, decorated)


def skip_for_integration(fn: _Markable) -> _Markable:
    decorated = pytest.mark.skipif(env.is_integration(), reason='Skipped in integration tests')(fn)
    return cast(_Markable, decorated)


def only_for_integration(fn: _Markable) -> _Markable:
    decorated = pytest.mark.skipif(not env.is_integration(), reason='Only for integration tests')(fn)
    return cast(_Markable, decorated)


def skip_for_unit(fn: _Markable) -> _Markable:
    decorated = pytest.mark.skipif(env.is_test() and not env.is_integration(), reason='Skipped in unit tests')(fn)
    return cast(_Markable, decorated)


def only_for_unit(fn: _Markable) -> _Markable:
    decorated = pytest.mark.skipif(not env.is_test(), reason='Only for unit tests')(fn)
    return cast(_Markable, decorated)



from invoke import Collection, Context, task


@task
def docs(c: Context):
    """Remove docs."""
    c.run('rm -rf docs')


@task
def python(c: Context):
    """Remove python artifacts."""
    c.run('find . -name "*.pyc" -delete')
    c.run('find . -name "__pycache__" -delete')
    c.run('rm -rf dist .nox')


@task
def coverage(c: Context):
    """Remove coverage files."""
    c.run('rm -rf coverage')


@task(docs, python, coverage)
def all(c: Context):  # noqa: A001, WPS125
    """Clean everything."""
    ...


ns = Collection(coverage, python, docs)
ns.add_task(all, default=True)



"""Helper to get/cache various variables to be used during invoke tasks.

Provides a mechansim to declare keys, and how to determine the value of those keys.
The values are only resolved once, and can refer to other values in a form of informal graph.

For example:

from outcome.devkit.invoke import env

# The most basic scenario is to declare a value
env.declare('my_key', 'my constant')


# Or you can declare a value based on a function. The name of the function will be the key.
@env.add
def my_key(e: env.Env) -> str:
    return 'my static value'


# Functions can refer to other env keys, creating a graph - be careful of loops!
@env.add
def my_graph_key(e: env.Env) -> str:
    return e.read('my_key')


# As a shortcut, you can just refer to an environment variable
env.from_os('MY_ENV_VAR')
env.from_os('my_alias', os_key='MY_ENV_VAR')


# Then, you can read the value wherever
def my_func():
    val = env.read('my_key')
"""

from __future__ import annotations

import os
from contextlib import contextmanager
from typing import Callable, Dict, Optional, Union, cast

SourceFn = Callable[['Env'], Optional[str]]
Source = Union[SourceFn, str]
ReadKey = Union[str, 'EnvItem']


class EnvItem:
    """An EnvItem represents a specific key and value, and how to calculate that value.

    You shouldn't have to explicitly access EnvItem objects.
    """

    name: str
    run: bool
    value: Optional[str]
    source: Source
    env: Env
    required: bool

    def __init__(self, name: str, source: Source, env: Env, required: bool) -> None:
        self.name = name
        self.run = False
        self.value = None
        self.source = source
        self.env = env
        self.required = required

    def read(self) -> Optional[str]:
        if not self.run:
            if callable(self.source):
                self.value = self.source(self.env)
            else:
                self.value = self.source

            self.run = True

        if self.required and self.value is None:
            raise RuntimeError(f"Env item '{self.name}' is required!")

        return self.value


@contextmanager
def as_required(e: EnvItem):
    previous = e.required
    e.required = True
    yield
    e.required = previous


class Env:  # noqa: WPS214
    items: Dict[str, EnvItem]

    def __init__(self):
        self.reset()

    def add(self, fn: SourceFn, required: bool = True, key: Optional[str] = None) -> EnvItem:
        key = key or fn.__name__
        return self._add(key, fn, required)

    def read(self, key: ReadKey, require: bool = False) -> Optional[str]:
        if isinstance(key, EnvItem):
            key = key.name

        try:
            item = self.items[key]

            if require:
                with as_required(item):
                    return item.read()
            else:
                return item.read()

        except KeyError:
            raise ValueError(f'Unknown env item! {key}')

    def strict_read(self, key: ReadKey) -> str:
        return cast(str, self.read(key, require=True))

    def from_os(self, key: str, os_key: Optional[str] = None, required: bool = True) -> EnvItem:
        effective_os_key = os_key or key

        def read_from_os(env: Env) -> Optional[str]:
            return os.environ.get(effective_os_key, None)

        return self.add(read_from_os, required=required, key=key)

    def declare(self, key: str, value: str) -> EnvItem:
        return self._add(key, value, required=True)

    def reset(self):
        self.items = {}

    def _add(self, key: str, source: Source, required: bool) -> EnvItem:
        if key in self.items:
            raise ValueError(f'Duplicate env item! {key}')

        self.items[key] = EnvItem(key, source, self, required)

        return self.items[key]


env = Env()


read = r = env.strict_read  # noqa: WPS429
add = env.add
from_os = env.from_os
declare = env.declare
reset = env.reset



from invoke import Collection
from outcome.devkit.invoke.tasks import check, clean, release, setup, test
from outcome.utils import env

namespace = Collection(setup, clean, release, check, test)

namespace.configure({'run': {'echo': True, 'pty': not env.is_ci()}})



"""A fixture to hide modules from import during tests.

For example, you want to test the following code:

```
try:
    import peewee
    HAS_DATABASE = True
except ImportError:
    HAS_DATABASE = False
```

You can write the following test:
```
@pytest.mark.parametrize('modules_to_hide', [['peewee']])
@pytest.mark.usefixtures('with_hidden_modules')
def test_has_no_peewee():
    assert not my_mod.HAS_DATABASE
"""

import sys
from contextlib import contextmanager
from typing import List

import pytest


@pytest.fixture
def modules_to_hide() -> List[str]:  # pragma: no cover
    return []


@pytest.fixture
def with_hidden_modules(modules_to_hide: List[str]):
    with _hide_modules(*modules_to_hide):
        yield


@contextmanager
def _hide_modules(*modules: str):
    # Get all currently loaded modules
    excluded_modules = [m for m in sys.modules.keys() if any(m.startswith(ex_m) for ex_m in modules)]

    extracted_modules = {}

    # Delete them from the sys.modules
    # Setting them to None triggers an ImportError on attempted
    # import
    for m in excluded_modules:
        extracted_modules[m] = sys.modules[m]
        sys.modules[m] = None  # type: ignore

    yield

    # Clear the None so we can re-import if necessary
    for n in excluded_modules:
        sys.modules[n] = extracted_modules[n]



"""Pact test helpers."""
from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, cast

from pactman import Pact as OriginalPact

if TYPE_CHECKING:  # pragma: no cover
    from pactman.mock.pact import V3Interaction  # noqa: WPS433


# This is a helper decorator that should be used on pact setup functions
# it automatically uses the pact context manager to in pytest fixtures
def with_pact(func: Callable[..., Pact]):  # pragma: no cover
    @wraps(func)
    def wrapped(*args: Any, **kwargs: Any):
        pact = func(*args, **kwargs)
        with pact:
            yield pact

    return wrapped


class Pact(OriginalPact):  # pragma: no cover
    # This is a fix - the official pactman library inserts new interactions at the beginning
    # of the _interactions array, but then uses the last item in the array when using `and_given`
    # This override addresses the index alignment issue.
    def and_given(self, provider_state: str, **params: str):

        self.semver

        if self.semver.major < 3:
            raise ValueError('pact v2 only allows a single provider state')
        elif not self._interactions:
            raise ValueError('only invoke and_given() after given()')

        # We know that we're dealing with V3
        most_recent_interaction = cast(V3Interaction, self._interactions[0])
        most_recent_interaction['providerStates'].append({'name': provider_state, 'params': params})
        return self


__all__ = ['Pact', 'with_pact']



from invoke import Collection, Context, task
from outcome.devkit.invoke.tasks import clean


@task(clean.all)
def build(c: Context):
    """Build the python package."""
    c.run('poetry build')


@task(build)
def publish(c: Context):
    """Publish the python package."""
    c.run('poetry publish')


ns = Collection(build, publish)



from typing import Optional

from invoke import Collection, Context, run, task
from outcome.devkit.invoke import env
from outcome.read_toml import lib as read_toml
from outcome.utils.env import is_ci

env.declare('pyproject.toml', './pyproject.toml')


@env.add
def build_system_requirements(e: env.Env) -> Optional[str]:
    return read_toml.read_from_file(env.r('pyproject.toml'), 'build-system.requires')


@env.add
def poetry_ld_flags(e: env.Env) -> str:
    # Retrieve the path to the ssl library, used for compiling some python C extensions
    path: str = run('brew --prefix openssl', echo=False, hide=True).stdout
    return f'{path}/lib'


@task
def build_system(c: Context):
    """Install essential build system components."""
    requirements = env.r(build_system_requirements)
    c.run(f'pip install {requirements}')

    if not is_ci():
        # Setup pre-commit
        c.run('pre-commit install -t pre-commit')
        c.run('pre-commit install -t pre-push')
        c.run('pre-commit install -t commit-msg')


@task(build_system)
def ci(c: Context):
    """Install the dependencies for CI environments."""
    c.run('poetry install --no-interaction --no-ansi --remove-untracked')


@task(build_system)
def dev(c: Context):
    """Install the dependencies for dev environments."""
    c.run('poetry install --remove-untracked', env={'LDFLAGS': env.r(poetry_ld_flags)})


@task(build_system)
def production(c: Context):
    """Install the dependencies for production environments."""
    c.run('poetry config virtualenvs.create false')
    c.run('poetry install --no-dev --no-interaction --no-ansi  --remove-untracked')


@task
def auto(c: Context):
    """Install either dev or CI dependencies, based on the environment."""
    if is_ci():
        ci(c)
    else:
        dev(c)


namespace = Collection(build_system, ci, dev, production, auto)



from pathlib import Path
from typing import Literal, Optional, Sequence, cast

from invoke import Collection, Context, Result, UnexpectedExit, task
from outcome.devkit.invoke import env
from outcome.devkit.invoke.tasks import clean
from outcome.utils.env import is_ci
from pydantic import BaseModel, Field
from rich.console import Console


class PyrightDiagnosticRangeBoundary(BaseModel):
    line: int
    character: int


class PyrightDiagnosticRange(BaseModel):
    start: PyrightDiagnosticRangeBoundary
    end: PyrightDiagnosticRangeBoundary


class PyrightDiagnostic(BaseModel):
    file: str
    severity: Literal['error', 'warning', 'information']
    message: str
    rule: Optional[str]
    range: PyrightDiagnosticRange  # noqa: A003, WPS125


class PyrightOutputSummary(BaseModel):
    error_count: int = Field(..., alias='errorCount')
    files_analyzed: int = Field(..., alias='filesAnalyzed')
    warning_count: int = Field(..., alias='warningCount')


class PyrightOutput(BaseModel):
    class Config:
        extra = 'ignore'

    diagnostics: Sequence[PyrightDiagnostic]
    summary: PyrightOutputSummary


@env.add
def code_dirs(e: env.Env) -> str:
    return ' '.join(d for d in ('./src', './bin', './test') if Path(d).is_dir())


@task(clean.all)
def types(c: Context, show_information: bool = False):
    """Run type-checking."""
    res = cast(Result, c.run(f'poetry run pyright {env.r(code_dirs)} --outputjson', echo=True, hide=True, warn=True))

    assert isinstance(res.stdout, str)
    results = PyrightOutput.parse_raw(res.stdout)

    console = Console()

    console.print(
        f'pyright analyzed {results.summary.files_analyzed} file(s), with {results.summary.warning_count} warnings and {results.summary.error_count} errors',  # noqa: E501
        style='cyan',
    )

    skipped_statuses = set()
    if not show_information:
        skipped_statuses.add('information')

    for d in (d for d in results.diagnostics if d.severity not in skipped_statuses):  # noqa: WPS335
        if d.severity == 'error':
            color = 'red'
        elif d.severity == 'warning':
            color = 'yellow'
        else:
            color = 'blue'

        message = f'[dim]({d.rule})[/dim] {d.message}' if d.rule else d.message
        console.print(
            f'{d.file}:{d.range.start.line+1}:{d.range.start.character+1} - [bold {color}]{d.severity}[/bold {color}] {message}',
        )

    if results.summary.error_count > 0:
        res.hide = ()
        raise UnexpectedExit(res, 'Error checking types')


@task(clean.all)
def format(c: Context):  # noqa: A001, WPS125
    """Run formatter."""
    if is_ci():
        c.run(f'poetry run black --check {env.r(code_dirs)}')
    else:
        c.run(f'poetry run black {env.r(code_dirs)}')


@task(clean.all)
def isort(c: Context):
    """Run isort."""
    if is_ci():
        c.run(f'poetry run isort -rc {env.r(code_dirs)} --check-only')
    else:
        c.run(f'poetry run isort -rc {env.r(code_dirs)}')


@task(clean.all)
def lint(c: Context):
    """Run flake8."""
    c.run(f'poetry run flake8 {env.r(code_dirs)}')


@task(clean.all, types, isort, format, lint)
def all(c: Context):  # noqa: A001, WPS125
    """Run all checks."""
    ...


ns = Collection(lint, isort, format, types)
ns.add_task(all, default=True)



from invoke import Collection, Context, task
from outcome.devkit.invoke.tasks import clean
from outcome.utils.env import is_ci

_default_target = './test'


def run_tests(  # noqa: WPS211
    c: Context, app_env: str, coverage: str, target: str, capture: bool = True, update_snapshots: bool = False,
):
    options = ['-vvv', '--ff', '--maxfail=1']

    if capture:
        options.append('-s')

    if update_snapshots:
        options.append('--snapshot-update')

    options_string = ' '.join(options)
    env = {'APP_ENV': app_env, 'PYTHONPATH': 'src'}

    # The in_stream is to play nice with pytest's capturing
    c.run(f'poetry run coverage run --context={coverage} -m pytest {options_string} {target}', env=env, in_stream=not capture)

    if is_ci():
        c.run('poetry run coverage report -m')
    else:
        # Run a first coverage, to the console but don't fail based on %
        c.run('poetry run coverage report -m --fail-under 0')
        # Run a second coverage to output html, and fail according to %
        c.run('poetry run coverage html --show-contexts')


@task(clean.all)
def unit(c: Context, target: str = _default_target, capture: bool = True, update_snapshots: bool = False):
    """Run unit tests."""
    run_tests(c, 'test', 'unit', target, capture, update_snapshots)


@task(clean.all)
def integration(c: Context, target: str = _default_target, capture: bool = True, update_snapshots: bool = False):
    """Run integration tests."""
    run_tests(c, 'integration', 'integration', target, capture, update_snapshots)


@task(clean.all, unit, integration)
def all(c: Context):  # noqa: A001, WPS125
    """Run unit and integration tests."""
    ...


ns = Collection(unit, integration)
ns.add_task(all, default=True)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['outcome',
 'outcome.devkit',
 'outcome.devkit.fixtures',
 'outcome.devkit.invoke',
 'outcome.devkit.invoke.tasks']

package_data = \
{'': ['*']}

install_requires = \
['black>=19.10b0,<20.0',
 'colored-traceback>=0.3.0,<0.4.0',
 'coverage>=5.0.3,<6.0.0',
 'flake8-breakpoint>=1.1.0,<2.0.0',
 'flake8-builtins>=1.5.2,<2.0.0',
 'flake8-colors>=0.1.6,<0.2.0',
 'flake8-commas>=2.0.0,<3.0.0',
 'flake8-debugger>=3.2.1,<4.0.0',
 'flake8-if-expr>=1.0.0,<2.0.0',
 'flake8-mutable>=1.2.0,<2.0.0',
 'flake8-print>=3.1.4,<5.0.0',
 'flake8-pytest>=1.3,<2.0',
 'flake8-return>=1.1.1,<2.0.0',
 'flake8>=3.7.9,<4.0.0',
 'ipython>=7.13.0,<8.0.0',
 'isort>=4.3.21,<5.0.0',
 'magicinvoke>=2.4.5,<3.0.0',
 'outcome-read-toml>=2.1.1,<3.0.0',
 'outcome-stubs>=0.3.0,<0.4.0',
 'outcome-utils>=5.0.0,<6.0.0',
 'pactman>=2.28.0,<3.0.0',
 'pdbpp>=0.10.2,<0.11.0',
 'poetry==1.1.4',
 'pre-commit>=2.10.1,<3.0.0',
 'pydantic>=1.7.3,<2.0.0',
 'pytest-asyncio>=0.12,<0.15',
 'pytest>=5.3.5,<7.0.0',
 'rich>=9.11.0,<10.0.0',
 'syrupy>=1.1.0,<2.0.0',
 'wemake-python-styleguide>=0.14.0,<0.15.0']

setup_kwargs = {
    'name': 'outcome-devkit',
    'version': '5.4.0',
    'description': 'A package containing common dev dependencies for python projects.',
    'long_description': '# devkit-py\n![Continuous Integration](https://github.com/outcome-co/devkit-py/workflows/Continuous%20Integration/badge.svg) ![version-badge](https://img.shields.io/badge/version-5.4.0-brightgreen)\n\nA package containing common dev dependencies for python projects.\n\n## Usage\n\n```sh\npoetry add -D outcome-devkit\n```\n\n## Development\n\nRun `./bootstrap.sh` when you clone the repository to get set up.\n',
    'author': 'Douglas Willcocks',
    'author_email': 'douglas@outcome.co',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/outcome-co/devkit-py',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.6,<4.0.0',
}


setup(**setup_kwargs)
